@@ -1,6 +1,6 @@
 /*******************************************************************************
 
-    uBlock Origin - a browser extension to block requests.
+    uBlock Origin - a comprehensive, efficient content blocker
     Copyright (C) 2014-present Raymond Hill
 
     This program is free software: you can redistribute it and/or modify
@@ -19,24 +19,19 @@
     Home: https://github.com/gorhill/uBlock
 */
 
-/* global CSS */
-
-'use strict';
-
-/******************************************************************************/
-/******************************************************************************/
-
 (async ( ) => {
 
 /******************************************************************************/
 
-const epickerId = vAPI.randomToken();
-let epickerConnectionId;
+if ( typeof vAPI !== 'object' ) { return; }
+if ( vAPI === null ) { return; }
 
-let pickerRoot = document.querySelector(`[${vAPI.sessionId}]`);
-if ( pickerRoot !== null ) { return; }
+if ( vAPI.pickerFrame ) { return; }
+vAPI.pickerFrame = true;
 
-let pickerBootArgs;
+const pickerUniqueId = vAPI.randomToken();
+
+const reCosmeticAnchor = /^#(\$|\?|\$\?)?#/;
 
 const netFilterCandidates = [];
 const cosmeticFilterCandidates = [];
@@ -75,17 +70,18 @@ const getElementBoundingClientRect = function(elem) {
     if ( rect.width !== 0 && rect.height !== 0 ) {
         return rect;
     }
+    if ( elem.shadowRoot instanceof DocumentFragment ) {
+        return getElementBoundingClientRect(elem.shadowRoot);
+    }
 
     let left = rect.left,
-        right = rect.right,
+        right = left + rect.width,
         top = rect.top,
-        bottom = rect.bottom;
+        bottom = top + rect.height;
 
     for ( const child of elem.children ) {
         rect = getElementBoundingClientRect(child);
-        if ( rect.width === 0 || rect.height === 0 ) {
-            continue;
-        }
+        if ( rect.width === 0 || rect.height === 0 ) { continue; }
         if ( rect.left < left ) { left = rect.left; }
         if ( rect.right > right ) { right = rect.right; }
         if ( rect.top < top ) { top = rect.top; }
@@ -93,8 +89,10 @@ const getElementBoundingClientRect = function(elem) {
     }
 
     return {
+        bottom,
         height: bottom - top,
         left,
+        right,
         top,
         width: right - left
     };
@@ -118,7 +116,7 @@ const highlightElements = function(elems, force) {
     const islands = [];
 
     for ( const elem of elems ) {
-        if ( elem === pickerRoot ) { continue; }
+        if ( elem === pickerFrame ) { continue; }
         targetElements.push(elem);
         const rect = getElementBoundingClientRect(elem);
         // Ignore offscreen areas
@@ -133,7 +131,7 @@ const highlightElements = function(elems, force) {
         );
     }
 
-    vAPI.MessagingConnection.sendTo(epickerConnectionId, {
+    pickerFramePort.postMessage({
         what: 'svgPaths',
         ocean: `M0 0h${ow}v${oh}h-${ow}z`,
         islands: islands.join(''),
@@ -174,7 +172,11 @@ const mergeStrings = function(urls) {
     merged = merged.replace(/^\*+$/, '')
                    .replace(/\*{2,}/g, '*')
                    .replace(/([^*]{1,3}\*)(?:[^*]{1,3}\*)+/g, '$1');
-    return merged;
+
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/1494
+    let pos = merged.indexOf('/');
+    if ( pos === -1 ) { pos = merged.length; }
+    return merged.slice(0, pos).includes('*') ? urls[0] : merged;
 };
 
 /******************************************************************************/
@@ -206,7 +208,8 @@ const backgroundImageURLFromElement = function(elem) {
 // https://github.com/gorhill/uBlock/issues/1725#issuecomment-226479197
 //   Limit returned string to 1024 characters.
 //   Also, return only URLs which will be seen by an HTTP observer.
-
+// https://github.com/uBlockOrigin/uBlock-issues/issues/2260
+//   Maybe get to the actual URL indirectly.
 const resourceURLsFromElement = function(elem) {
     const urls = [];
     const tagName = elem.localName;
@@ -216,13 +219,15 @@ const resourceURLsFromElement = function(elem) {
         if ( url !== '' ) { urls.push(url); }
         return urls;
     }
-    {
-        const s = elem[prop];
-        if ( typeof s === 'string' && /^https?:\/\//.test(s) ) {
-            urls.push(trimFragmentFromURL(s.slice(0, 1024)));
-        }
+    let s = elem[prop];
+    if ( s instanceof SVGAnimatedString ) {
+        s = s.baseVal;
+    }
+    if ( typeof s === 'string' && /^https?:\/\//.test(s) ) {
+        urls.push(trimFragmentFromURL(s.slice(0, 1024)));
     }
     resourceURLsFromSrcset(elem, urls);
+    resourceURLsFromPicture(elem, urls);
     return urls;
 };
 
@@ -257,6 +262,20 @@ const resourceURLsFromSrcset = function(elem, out) {
     }
 };
 
+// https://github.com/uBlockOrigin/uBlock-issues/issues/2069#issuecomment-1080600661
+// https://developer.mozilla.org/en-US/docs/Web/HTML/Element/picture
+const resourceURLsFromPicture = function(elem, out) {
+    if ( elem.localName === 'source' ) { return; }
+    const picture = elem.parentElement;
+    if ( picture === null || picture.localName !== 'picture' ) { return; }
+    const sources = picture.querySelectorAll(':scope > source');
+    for ( const source of sources ) {
+        const urls = resourceURLsFromElement(source);
+        if ( urls.length === 0 ) { continue; }
+        out.push(...urls);
+    }
+};
+
 /******************************************************************************/
 
 const netFilterFromUnion = function(patternIn, out) {
@@ -321,7 +340,7 @@ const netFilterFromElement = function(elem) {
     const pattern = mergeStrings(urls);
 
 
-    if ( bestCandidateFilter === null ) {
+    if ( bestCandidateFilter === null && elem.matches('html,body') === false ) {
         bestCandidateFilter = {
             type: 'net',
             filters: candidates,
@@ -348,7 +367,9 @@ const netFilter1stSources = {
      'embed': 'src',
     'iframe': 'src',
        'img': 'src',
+     'image': 'href',
     'object': 'data',
+    'source': 'src',
      'video': 'src'
 };
 
@@ -372,6 +393,7 @@ const filterTypes = {
 const cosmeticFilterFromElement = function(elem) {
     if ( elem === null ) { return 0; }
     if ( elem.nodeType !== 1 ) { return 0; }
+    if ( noCosmeticFiltering ) { return 0; }
 
     if ( candidateElements.indexOf(elem) === -1 ) {
         candidateElements.push(elem);
@@ -395,11 +417,14 @@ const cosmeticFilterFromElement = function(elem) {
     }
 
     // Tag name
-    const tagName = elem.localName;
+    const tagName = CSS.escape(elem.localName);
 
     // Use attributes if still no selector found.
     // https://github.com/gorhill/uBlock/issues/1901
     //   Trim attribute value, this may help in case of malformed HTML.
+    //
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/1923
+    //   Escape unescaped `"` in attribute values
     if ( selector === '' ) {
         let attributes = [], attr;
         switch ( tagName ) {
@@ -441,13 +466,14 @@ const cosmeticFilterFromElement = function(elem) {
         }
         while ( (attr = attributes.pop()) ) {
             if ( attr.v.length === 0 ) { continue; }
+            const w = attr.v.replace(/([^\\])"/g, '$1\\"');
             v = elem.getAttribute(attr.k);
             if ( attr.v === v ) {
-                selector += `[${attr.k}="${attr.v}"]`;
+                selector += `[${attr.k}="${w}"]`;
             } else if ( v.startsWith(attr.v) ) {
-                selector += `[${attr.k}^="${attr.v}"]`;
+                selector += `[${attr.k}^="${w}"]`;
             } else {
-                selector += `[${attr.k}*="${attr.v}"]`;
+                selector += `[${attr.k}*="${w}"]`;
             }
         }
     }
@@ -511,8 +537,19 @@ const filtersFrom = function(x, y) {
         x = undefined;
     }
 
-    // Network filter from element which was clicked.
-    if ( first !== null ) {
+    // https://github.com/gorhill/uBlock/issues/1545
+    //   Network filter candidates from all other elements found at [x,y].
+    // https://www.reddit.com/r/uBlockOrigin/comments/qmjk36/
+    //   Extract network candidates first.
+    if ( typeof x === 'number' ) {
+        const magicAttr = `${pickerUniqueId}-clickblind`;
+        pickerFrame.setAttribute(magicAttr, '');
+        const elems = document.elementsFromPoint(x, y);
+        pickerFrame.removeAttribute(magicAttr);
+        for ( const elem of elems ) {
+            netFilterFromElement(elem);
+        }
+    } else if ( first !== null ) {
         netFilterFromElement(first);
     }
 
@@ -535,24 +572,16 @@ const filtersFrom = function(x, y) {
         }
     }
 
-    // https://github.com/gorhill/uBlock/issues/1545
-    //   Network filter candidates from all other elements found at
-    //   point (x, y).
-    if ( typeof x === 'number' ) {
-        const attrName = vAPI.sessionId + '-clickblind';
-        elem = first;
-        while ( elem !== null ) {
-            const previous = elem;
-            elem.setAttribute(attrName, '');
-            elem = elementFromPoint(x, y);
-            if ( elem === null || elem === previous ) { break; }
-            netFilterFromElement(elem);
-        }
-        for ( const elem of document.querySelectorAll(`[${attrName}]`) ) {
-            elem.removeAttribute(attrName);
-        }
-
-        netFilterFromElement(document.body);
+    // https://github.com/gorhill/uBlock/commit/ebaa8a8bb28aef043a68c99965fe6c128a3fe5e4#commitcomment-63818019
+    //   If still no best candidate, just use whatever is available in network
+    //   filter candidates -- which may have been previously skipped in favor
+    //   of cosmetic filters.
+    if ( bestCandidateFilter === null && netFilterCandidates.length !== 0 ) {
+        bestCandidateFilter = {
+            type: 'net',
+            filters: netFilterCandidates,
+            slot: 0
+        };
     }
 
     return netFilterCandidates.length + cosmeticFilterCandidates.length;
@@ -583,6 +612,21 @@ const filterToDOMInterface = (( ) => {
     const reCaret = '(?:[^%.0-9a-z_-]|$)';
     const rePseudoElements = /:(?::?after|:?before|:[a-z-]+)$/;
 
+    const matchElemToRegex = (elem, re) => {
+        const srcProp = netFilter1stSources[elem.localName];
+        let src = elem[srcProp];
+        if ( src instanceof SVGAnimatedString ) {
+            src = src.baseVal;
+        }
+        if ( typeof src === 'string' && /^https?:\/\//.test(src) ) {
+            if ( re.test(src) ) { return srcProp; }
+        }
+        src = elem.currentSrc;
+        if ( typeof src === 'string' && /^https?:\/\//.test(src) ) {
+            if ( re.test(src) ) { return srcProp; }
+        }
+    };
+
     // Net filters: we need to lookup manually -- translating into a foolproof
     // CSS selector is just not possible.
     //
@@ -634,25 +678,23 @@ const filterToDOMInterface = (( ) => {
         }
 
         // Lookup by tag names.
+        // https://github.com/uBlockOrigin/uBlock-issues/issues/2260
+        //   Maybe get to the actual URL indirectly.
+        //
+        // https://github.com/uBlockOrigin/uBlock-issues/issues/3142
+        //   Don't try to match against non-network URIs.
         const elems = document.querySelectorAll(
             Object.keys(netFilter1stSources).join()
         );
         for ( const elem of elems ) {
-            const srcProp = netFilter1stSources[elem.localName];
-            const src = elem[srcProp];
-            if (
-                typeof src === 'string' &&
-                    reFilter.test(src) ||
-                typeof elem.currentSrc === 'string' &&
-                    reFilter.test(elem.currentSrc)
-            ) {
-                out.push({
-                    elem,
-                    src: srcProp,
-                    opt: filterTypes[elem.localName],
-                    style: vAPI.hideStyle,
-                });
-            }
+            const srcProp = matchElemToRegex(elem, reFilter);
+            if ( srcProp === undefined ) { continue; }
+            out.push({
+                elem,
+                src: srcProp,
+                opt: filterTypes[elem.localName],
+                style: vAPI.hideStyle,
+            });
         }
 
         // Find matching background image in current set of candidate elements.
@@ -691,7 +733,7 @@ const filterToDOMInterface = (( ) => {
         }
         const out = [];
         for ( const elem of elems ) {
-            if ( elem === pickerRoot ) { continue; }
+            if ( elem === pickerFrame ) { continue; }
             out.push({ elem, raw, style: vAPI.hideStyle });
         }
         return out;
@@ -703,14 +745,23 @@ const filterToDOMInterface = (( ) => {
     // https://github.com/gorhill/uBlock/issues/2515
     //   Remove trailing pseudo-element when querying.
     const fromCompiledCosmeticFilter = function(raw) {
+        if ( noCosmeticFiltering ) { return; }
         if ( typeof raw !== 'string' ) { return; }
         let elems, style;
         try {
             const o = JSON.parse(raw);
             elems = vAPI.domFilterer.createProceduralFilter(o).exec();
-            style = o.action === undefined || o.action[0] !== ':style'
-                ? vAPI.hideStyle
-                : o.action[1];
+            switch ( o.action && o.action[0] || '' ) {
+            case '':
+            case 'remove':
+                style = vAPI.hideStyle;
+                break;
+            case 'style':
+                style = o.action[1];
+                break;
+            default:
+                break;
+            }
         } catch(ex) {
             return;
         }
@@ -739,7 +790,7 @@ const filterToDOMInterface = (( ) => {
             return;
         }
         lastFilter = filter;
-        if ( filter.startsWith('##') === false ) {
+        if ( reCosmeticAnchor.test(filter) === false ) {
             lastResultset = fromNetworkFilter(filter);
             if ( previewing ) { apply(); }
             return lastResultset;
@@ -760,7 +811,8 @@ const filterToDOMInterface = (( ) => {
         if ( Array.isArray(lastResultset) === false ) { return; }
         const rootElem = document.documentElement;
         for ( const { elem, style } of lastResultset ) {
-            if ( elem === pickerRoot ) { continue; }
+            if ( elem === pickerFrame ) { continue; }
+            if ( style === undefined ) { continue; }
             if ( elem === rootElem && style === vAPI.hideStyle ) { continue; }
             let styleToken = vAPI.epickerStyleProxies.get(style);
             if ( styleToken === undefined ) {
@@ -789,10 +841,10 @@ const filterToDOMInterface = (( ) => {
             return unapply();
         }
         if ( Array.isArray(lastResultset) === false ) { return; }
-        if ( permanent === false || lastFilter.startsWith('##') === false ) {
+        if ( permanent === false || reCosmeticAnchor.test(lastFilter) === false ) {
             return apply();
         }
-        if ( vAPI.domFilterer instanceof Object === false ) { return; }
+        if ( noCosmeticFiltering ) { return; }
         const cssSelectors = new Set();
         const proceduralSelectors = new Set();
         for ( const { raw } of lastResultset ) {
@@ -803,9 +855,9 @@ const filterToDOMInterface = (( ) => {
             }
         }
         if ( cssSelectors.size !== 0 ) {
-            vAPI.domFilterer.addCSSRule(
-                Array.from(cssSelectors),
-                vAPI.hideStyle
+            vAPI.domFilterer.addCSS(
+                `${Array.from(cssSelectors).join('\n')}\n{${vAPI.hideStyle}}`,
+                { mustInject: true }
             );
         }
         if ( proceduralSelectors.size !== 0 ) {
@@ -820,11 +872,43 @@ const filterToDOMInterface = (( ) => {
 
 /******************************************************************************/
 
+const onOptimizeCandidates = function(details) {
+    const { candidates } = details;
+    const results = [];
+    for ( const paths of candidates ) {
+        let count = Number.MAX_SAFE_INTEGER;
+        let selector = '';
+        for ( let i = 0, n = paths.length; i < n; i++ ) {
+            const s = paths.slice(n - i - 1).join('');
+            const elems = document.querySelectorAll(s);
+            if ( elems.length < count ) {
+                selector = s;
+                count = elems.length;
+            }
+        }
+        results.push({ selector: `##${selector}`, count });
+    }
+    // Sort by most match count and shortest selector to least match count and
+    // longest selector.
+    results.sort((a, b) => {
+        const r = b.count - a.count;
+        if ( r !== 0 ) { return r; }
+        return a.selector.length - b.selector.length;
+    });
+
+    pickerFramePort.postMessage({
+        what: 'candidatesOptimized',
+        candidates: results.map(a => a.selector),
+        slot: details.slot,
+    });
+};
+
+/******************************************************************************/
+
 const showDialog = function(options) {
-    vAPI.MessagingConnection.sendTo(epickerConnectionId, {
+    pickerFramePort.postMessage({
         what: 'showDialog',
-        hostname: self.location.hostname,
-        origin: self.location.origin,
+        url: self.location.href,
         netFilters: netFilterCandidates,
         cosmeticFilters: cosmeticFilterCandidates,
         filter: bestCandidateFilter,
@@ -845,15 +929,23 @@ const elementFromPoint = (( ) => {
         } else {
             return null;
         }
-        if ( !pickerRoot ) { return null; }
-        const magicAttr = `${vAPI.sessionId}-clickblind`;
-        pickerRoot.setAttribute(magicAttr, '');
+        if ( !pickerFrame ) { return null; }
+        const magicAttr = `${pickerUniqueId}-clickblind`;
+        pickerFrame.setAttribute(magicAttr, '');
         let elem = document.elementFromPoint(x, y);
-        if ( elem === document.body || elem === document.documentElement ) {
+        if (
+            elem === null || /* to skip following tests */
+            elem === document.body ||
+            elem === document.documentElement || (
+                pickerBootArgs.zap !== true &&
+                noCosmeticFiltering &&
+                resourceURLsFromElement(elem).length === 0
+            )
+        ) {
             elem = null;
         }
         // https://github.com/uBlockOrigin/uBlock-issues/issues/380
-        pickerRoot.removeAttribute(magicAttr);
+        pickerFrame.removeAttribute(magicAttr);
         return elem;
     };
 })();
@@ -889,36 +981,45 @@ const zapElementAtPoint = function(mx, my, options) {
         return;
     }
 
-    let elem = targetElements.length !== 0 && targetElements[0] || null;
-    if ( elem === null && mx !== undefined ) {
-        elem = elementFromPoint(mx, my);
+    let elemToRemove = targetElements.length !== 0 && targetElements[0] || null;
+    if ( elemToRemove === null && mx !== undefined ) {
+        elemToRemove = elementFromPoint(mx, my);
     }
 
-    if ( elem instanceof Element === false ) { return; }
+    if ( elemToRemove instanceof Element === false ) { return; }
 
-    const getStyleValue = function(elem, prop) {
+    const getStyleValue = (elem, prop) => {
         const style = window.getComputedStyle(elem);
         return style ? style[prop] : '';
     };
 
     // Heuristic to detect scroll-locking: remove such lock when detected.
-    if (
-        parseInt(getStyleValue(elem, 'zIndex'), 10) >= 1000 ||
-        getStyleValue(elem, 'position') === 'fixed'
-    ) {
+    let maybeScrollLocked = elemToRemove.shadowRoot instanceof DocumentFragment;
+    if ( maybeScrollLocked === false ) {
+        let elem = elemToRemove;
+        do {
+            maybeScrollLocked =
+                parseInt(getStyleValue(elem, 'zIndex'), 10) >= 1000 ||
+                getStyleValue(elem, 'position') === 'fixed';
+            elem = elem.parentElement;
+        } while ( elem !== null && maybeScrollLocked === false );
+    }
+    if ( maybeScrollLocked ) {
         const doc = document;
         if ( getStyleValue(doc.body, 'overflowY') === 'hidden' ) {
             doc.body.style.setProperty('overflow', 'auto', 'important');
         }
         if ( getStyleValue(doc.body, 'position') === 'fixed' ) {
-            doc.body.style.setProperty('position', 'static', 'important');
+            doc.body.style.setProperty('position', 'initial', 'important');
+        }
+        if ( getStyleValue(doc.documentElement, 'position') === 'fixed' ) {
+            doc.documentElement.style.setProperty('position', 'initial', 'important');
         }
         if ( getStyleValue(doc.documentElement, 'overflowY') === 'hidden' ) {
             doc.documentElement.style.setProperty('overflow', 'auto', 'important');
         }
     }
-
-    elem.remove();
+    elemToRemove.remove();
     highlightElementAtPoint(mx, my);
 };
 
@@ -960,7 +1061,7 @@ const onViewportChanged = function() {
 // Auto-select a specific target, if any, and if possible
 
 const startPicker = function() {
-    pickerRoot.focus();
+    pickerFrame.focus();
 
     self.addEventListener('scroll', onViewportChanged, { passive: true });
     self.addEventListener('resize', onViewportChanged, { passive: true });
@@ -969,6 +1070,7 @@ const startPicker = function() {
     // Try using mouse position
     if (
         pickerBootArgs.mouse &&
+        vAPI.mouseClick instanceof Object &&
         typeof vAPI.mouseClick.x === 'number' &&
         vAPI.mouseClick.x > 0
     ) {
@@ -996,15 +1098,29 @@ const startPicker = function() {
     if ( attr === undefined ) { return; }
     const elems = document.getElementsByTagName(tagName);
     for ( const elem of elems  ) {
-        if ( elem === pickerRoot ) { continue; }
-        const src = elem[attr];
-        if ( typeof src !== 'string' ) { continue; }
-        if ( (src !== url) && (src !== '' || url !== 'about:blank') ) {
+        if ( elem === pickerFrame ) { continue; }
+        const srcs = resourceURLsFromElement(elem);
+        if (
+            (srcs.length !== 0 && srcs.includes(url) === false) ||
+            (srcs.length === 0 && url !== 'about:blank')
+        ) {
             continue;
         }
-        elem.scrollIntoView({ behavior: 'smooth', block: 'start' });
         filtersFrom(elem);
-        return showDialog({ broad: true });
+        if (
+            netFilterCandidates.length !== 0 ||
+            cosmeticFilterCandidates.length !== 0
+        ) {
+            if ( pickerBootArgs.mouse !== true ) {
+                elem.scrollIntoView({
+                    behavior: 'smooth',
+                    block: 'center',
+                    inline: 'center'
+                });
+            }
+            showDialog({ broad: true });
+        }
+        return;
     }
 
     // A target was specified, but it wasn't found: abort.
@@ -1021,89 +1137,79 @@ const quitPicker = function() {
     self.removeEventListener('resize', onViewportChanged, { passive: true });
     self.removeEventListener('keydown', onKeyPressed, true);
     vAPI.shutdown.remove(quitPicker);
-    vAPI.MessagingConnection.disconnectFrom(epickerConnectionId);
-    vAPI.MessagingConnection.removeListener(onConnectionMessage);
+    if ( pickerFramePort ) {
+        pickerFramePort.close();
+        pickerFramePort = null;
+    }
+    if ( pickerFrame ) {
+        pickerFrame.remove();
+        pickerFrame = null;
+    }
     vAPI.userStylesheet.remove(pickerCSS);
     vAPI.userStylesheet.apply();
-
-    if ( pickerRoot === null ) { return; }
-
-    pickerRoot.remove();
-    pickerRoot = null;
-
+    vAPI.pickerFrame = false;
     self.focus();
 };
 
+vAPI.shutdown.add(quitPicker);
+
 /******************************************************************************/
 
 const onDialogMessage = function(msg) {
     switch ( msg.what ) {
-        case 'start':
-            startPicker();
-            if ( targetElements.length === 0 ) {
-                highlightElements([], true);
-            }
-            break;
-        case 'dialogCreate':
-            filterToDOMInterface.queryAll(msg);
-            filterToDOMInterface.preview(true, true);
-            quitPicker();
-            break;
-        case 'dialogSetFilter': {
-            const resultset = filterToDOMInterface.queryAll(msg) || [];
-            highlightElements(resultset.map(a => a.elem), true);
-            if ( msg.filter === '!' ) { break; }
-            vAPI.MessagingConnection.sendTo(epickerConnectionId, {
-                what: 'resultsetDetails',
-                count: resultset.length,
-                opt: resultset.length !== 0 ? resultset[0].opt : undefined,
-            });
-            break;
+    case 'start':
+        startPicker();
+        if ( pickerFramePort === null ) { break; }
+        if ( targetElements.length === 0 ) {
+            highlightElements([], true);
         }
-        case 'quitPicker':
-            filterToDOMInterface.preview(false);
-            quitPicker();
-            break;
-        case 'highlightElementAtPoint':
-            highlightElementAtPoint(msg.mx, msg.my);
-            break;
-        case 'unhighlight':
-            highlightElements([]);
-            break;
-        case 'filterElementAtPoint':
-            filterElementAtPoint(msg.mx, msg.my, msg.broad);
-            break;
-        case 'zapElementAtPoint':
-            zapElementAtPoint(msg.mx, msg.my, msg.options);
-            if ( msg.options.highlight !== true && msg.options.stay !== true ) {
-                quitPicker();
-            }
-            break;
-        case 'togglePreview':
-            filterToDOMInterface.preview(msg.state);
-            if ( msg.state === false ) {
-                highlightElements(targetElements, true);
-            }
-            break;
-        default:
-            break;
+        break;
+    case 'optimizeCandidates':
+        onOptimizeCandidates(msg);
+        break;
+    case 'dialogCreate':
+        filterToDOMInterface.queryAll(msg);
+        filterToDOMInterface.preview(true, true);
+        quitPicker();
+        break;
+    case 'dialogSetFilter': {
+        const resultset = filterToDOMInterface.queryAll(msg) || [];
+        highlightElements(resultset.map(a => a.elem), true);
+        if ( msg.filter === '!' ) { break; }
+        pickerFramePort.postMessage({
+            what: 'resultsetDetails',
+            count: resultset.length,
+            opt: resultset.length !== 0 ? resultset[0].opt : undefined,
+        });
+        break;
     }
-};
-
-/******************************************************************************/
-
-const onConnectionMessage = function(msg) {
-    if ( msg.from !== `epickerDialog-${epickerId}` ) { return; }
-    switch ( msg.what ) {
-        case 'connectionRequested':
-            epickerConnectionId = msg.id;
-            return true;
-        case 'connectionBroken':
+    case 'quitPicker':
+        filterToDOMInterface.preview(false);
+        quitPicker();
+        break;
+    case 'highlightElementAtPoint':
+        highlightElementAtPoint(msg.mx, msg.my);
+        break;
+    case 'unhighlight':
+        highlightElements([]);
+        break;
+    case 'filterElementAtPoint':
+        filterElementAtPoint(msg.mx, msg.my, msg.broad);
+        break;
+    case 'zapElementAtPoint':
+        zapElementAtPoint(msg.mx, msg.my, msg.options);
+        if ( msg.options.highlight !== true && msg.options.stay !== true ) {
             quitPicker();
-            break;
-        case 'connectionMessage':
-            onDialogMessage(msg.payload);
-            break;
+        }
+        break;
+    case 'togglePreview':
+        filterToDOMInterface.preview(msg.state);
+        if ( msg.state === false ) {
+            highlightElements(targetElements, true);
+        }
+        break;
+    default:
+        break;
     }
 };
 
@@ -1126,32 +1232,10 @@ const onConnectionMessage = function(msg) {
 // of the iframe, and cannot interfere with its style properties. However the
 // page can remove the iframe.
 
-// We need extra messaging capabilities + fetch/process picker arguments.
-{
-    const results = await Promise.all([
-        vAPI.messaging.extend(),
-        vAPI.messaging.send('elementPicker', { what: 'elementPickerArguments' }),
-    ]);
-    if ( results[0] !== true ) { return; }
-    pickerBootArgs = results[1];
-    if ( typeof pickerBootArgs !== 'object' || pickerBootArgs === null ) {
-        return;
-    }
-    // Restore net filter union data if origin is the same.
-    const eprom = pickerBootArgs.eprom || null;
-    if ( eprom !== null && eprom.lastNetFilterSession === lastNetFilterSession ) {
-        lastNetFilterHostname = eprom.lastNetFilterHostname || '';
-        lastNetFilterUnion = eprom.lastNetFilterUnion || '';
-    }
-}
-
 // The DOM filterer will not be present when cosmetic filtering is disabled.
-if (
-    pickerBootArgs.zap !== true &&
-    vAPI.domFilterer instanceof Object === false
-) {
-    return;
-}
+const noCosmeticFiltering =
+    vAPI.domFilterer instanceof Object === false ||
+    vAPI.noSpecificCosmeticFiltering === true;
 
 // https://github.com/gorhill/uBlock/issues/1529
 //   In addition to inline styles, harden the element picker styles by using
@@ -1161,8 +1245,11 @@ const pickerCSSStyle = [
     'border: 0',
     'border-radius: 0',
     'box-shadow: none',
+    'color-scheme: light dark',
     'display: block',
+    'filter: none',
     'height: 100vh',
+    '    height: 100svh',
     'left: 0',
     'margin: 0',
     'max-height: none',
@@ -1175,17 +1262,22 @@ const pickerCSSStyle = [
     'pointer-events: auto',
     'position: fixed',
     'top: 0',
-    'visibility: visible',
+    'transform: none',
+    'visibility: hidden',
     'width: 100%',
     'z-index: 2147483647',
     ''
-].join(' !important;');
+].join(' !important;\n');
+
 
 const pickerCSS = `
-:root > [${vAPI.sessionId}] {
+:root > [${pickerUniqueId}] {
     ${pickerCSSStyle}
 }
-:root [${vAPI.sessionId}-clickblind] {
+:root > [${pickerUniqueId}-loaded] {
+    visibility: visible !important;
+}
+:root [${pickerUniqueId}-clickblind] {
     pointer-events: none !important;
 }
 `;
@@ -1193,21 +1285,53 @@ const pickerCSS = `
 vAPI.userStylesheet.add(pickerCSS);
 vAPI.userStylesheet.apply();
 
-pickerRoot = document.createElement('iframe');
-pickerRoot.setAttribute(vAPI.sessionId, '');
-document.documentElement.append(pickerRoot);
+let pickerBootArgs;
+let pickerFramePort = null;
 
-vAPI.shutdown.add(quitPicker);
-
-vAPI.MessagingConnection.addListener(onConnectionMessage);
-
-{
+const bootstrap = async ( ) => {
+    pickerBootArgs = await vAPI.messaging.send('elementPicker', {
+        what: 'elementPickerArguments',
+    });
+    if ( typeof pickerBootArgs !== 'object' ) { return; }
+    if ( pickerBootArgs === null ) { return; }
+    // Restore net filter union data if origin is the same.
+    const eprom = pickerBootArgs.eprom || null;
+    if ( eprom !== null && eprom.lastNetFilterSession === lastNetFilterSession ) {
+        lastNetFilterHostname = eprom.lastNetFilterHostname || '';
+        lastNetFilterUnion = eprom.lastNetFilterUnion || '';
+    }
     const url = new URL(pickerBootArgs.pickerURL);
-    url.searchParams.set('epid', epickerId);
     if ( pickerBootArgs.zap ) {
         url.searchParams.set('zap', '1');
     }
-    pickerRoot.contentWindow.location = url.href;
+    return new Promise(resolve => {
+        const iframe = document.createElement('iframe');
+        iframe.setAttribute(pickerUniqueId, '');
+        document.documentElement.append(iframe);
+        iframe.addEventListener('load', ( ) => {
+            iframe.setAttribute(`${pickerUniqueId}-loaded`, '');
+            const channel = new MessageChannel();
+            pickerFramePort = channel.port1;
+            pickerFramePort.onmessage = ev => {
+                onDialogMessage(ev.data || {});
+            };
+            pickerFramePort.onmessageerror = ( ) => {
+                quitPicker();
+            };
+            iframe.contentWindow.postMessage(
+                { what: 'epickerStart' },
+                url.href,
+                [ channel.port2 ]
+            );
+            resolve(iframe);
+        }, { once: true });
+        iframe.contentWindow.location = url.href;
+    });
+};
+
+let pickerFrame = await bootstrap();
+if ( Boolean(pickerFrame) === false ) {
+    quitPicker();
 }
 
 /******************************************************************************/
